#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <stdio.h>
#include <stdlib.h>

#include "lcd.h"
#include "usart.h"

// frequency detector's variables

typedef unsigned char byte;
//data storage variables
byte newData = 0;
byte prevData = 0;
unsigned int time = 0;//keeps time and sends vales to store in timer[] occasionally
int timer[10];//storage for timing of events
int slope[10];//storage for slope of events
unsigned int totalTimer;//used to calculate period
unsigned int period;//storage for period of wave
byte index = 0;//current storage index
float frequency;//storage for frequency calculations
int maxSlope = 0;//used to calculate max slope as trigger point
int newSlope;//storage for incoming slope data

//variables for deciding whether you have a match
byte noMatch = 0;//counts how many non-matches you've received to reset variables if it's been too long
byte slopeTol = 3;//slope tolerance- adjust this if you need
int timerTol = 1;//timer tolerance- adjust this if you need

//variables for amp detection
unsigned int ampTimer = 0;
byte maxAmp = 0;
byte checkMaxAmp;
byte ampThreshold = 30;//30;//raise if you have a very noisy signal

// rest of variables
#define VREF 2.56
#define PRECISION 0.55
#define STRING_TOLERANCE 0.30
#define NUM_TUNINGS 3

static float standardE[] = {82.41, 110.0, 146.83, 196.0, 246.94, 329.63};
static char descriptionE[] = "Standard E";
static char stringsE[] =  "E  A  D  G  B  e";

static float dropD[] = {73.42, 110.0, 146.83, 196.0, 246.94, 329.63};
static char descriptionDropD[] = "  Drop D";
static char stringsDropD[] =  "D  A  D  G  B  e";

static float standardD[] = {73.42, 97.99, 130.81, 174.61, 220.0, 293.66};
static char descriptioStandardD[] = "Standard D";
static char stringsStandardD[] =  "D  G  C  F  A  d";

static float multiplier[] = {4, 4, 4, 4, 3, 2}; // speed multipliers are different for each string

int selected_tuning;
int selection_index;
float selected_string;

enum direction{FWD, BWD};
char reverse_directions;

int max_motor_turns;
int num_pulses;


void reset(){//clean out some variables
    index = 0;//reset index
    noMatch = 0;//reset match couner
    maxSlope = 0;//reset slope
}

// ADC conversion complete handler
ISR(ADC_vect)
{
    prevData = newData;//store previous value
    newData = ADCH;//get value from A0
    if (prevData < 127 && newData >=127){//if increasing and crossing midpoint
        newSlope = newData - prevData;//calculate slope
        if (abs(newSlope-maxSlope)<slopeTol){//if slopes are ==
            //record new data and reset time
            slope[index] = newSlope;
            timer[index] = time;
            time = 0;
            if (index == 0){//new max slope just reset
                noMatch = 0;
                index++;//increment index
            }
            else if (abs(timer[0]-timer[index])<timerTol && abs(slope[0]-newSlope)<slopeTol){//if timer duration and slopes match
                //sum timer values
                totalTimer = 0;
                for (byte i=0;i<index;i++){
                    totalTimer+=timer[i];
                }
                period = totalTimer;//set period
                //reset new zero index values to compare with
                timer[0] = timer[index];
                slope[0] = slope[index];
                index = 1;//set index to 1
                noMatch = 0;
            }
            else{//crossing midpoint but not match
                index++;//increment index
                if (index > 9){
                    reset();
                }
            }
        }
        else if (newSlope>maxSlope){//if new slope is much larger than max slope
            maxSlope = newSlope;
            time = 0;//reset clock
            noMatch = 0;
            index = 0;//reset index
        }
        else{//slope not steep enough
            noMatch++;//increment no match counter
            if (noMatch>9){
                reset();
            }
        }
    }

    time++;//increment timer at rate of 38.5kHz

    ampTimer++;//increment amplitude timer
    if (abs(127-ADCH)>maxAmp){
        maxAmp = abs(127-ADCH);
    }
    if (ampTimer==1000){
        ampTimer = 0;
        checkMaxAmp = maxAmp;
        maxAmp = 0;
    }
}

void init_adc() {
    // Initialize ADC
    ADCSRA = 0;
    ADCSRB = 0;

    ADMUX |= (1 << REFS1) | (1 << REFS0); //set reference voltage
    ADMUX |= (1 << ADLAR); //left align the ADC value- so we can read highest 8 bits from ADCH register only

    ADCSRA |= (1 << ADPS2) | (1 << ADPS0); //set ADC clock with 32 prescaler- 16mHz/32=500kHz
    ADCSRA |= (1 << ADATE); //enabble auto trigger
    ADCSRA |= (1 << ADIE); //enable interrupts when measurement complete
    ADCSRA |= (1 << ADEN); //enable ADC
    ADCSRA |= (1 << ADSC); //start ADC measurements
}

void init_PWM(){
    // intialise PWM motor control on OC0A

    //Set OC0 PIN as output. It is  PB3 on ATmega16 ATmega32
    DDRB|=(1<<PB3);

    TCCR0A |= (1<<WGM00)|(1<<WGM01)|(1<<COM0A1);
    TCCR0B |= (1<<CS01)|(1<<CS00);//|(1<<CS02);

    // set motor driver mode and direction to 1
    DDRB |= (1 << PB1) | (1 << PB2);
    PORTB |= (1 << PB1) | (1 << PB2);
}

void move_motor(enum direction d, unsigned char speed, int num_turns){
    // signal direction change to the motor driver

    if(d == FWD && reverse_directions){
        PORTB &= ~(1 << PB2);
    } else if (d == FWD && !reverse_directions){
        PORTB |= (1 << PB2);
    } else if( d == BWD && reverse_directions){
        PORTB |= (1 << PB2);
    } else if (d == BWD && !reverse_directions){
        PORTB &= ~(1 << PB2);
    } else {
        return;
    }

    // signal to the motor monitor to stop the motor after this number of turns
    max_motor_turns = num_turns;

    TCCR0A |= (1<<COM0A1); // enable pwm mode

    // set PWM signal duty cycle
    OCR0A = speed;
}

void stop_motor(){
    // turn off PWM on the driver control pin
    TCCR0A &= ~(1<<COM0A1);
    max_motor_turns = 0;

    frequency = 0;
}

ISR(PCINT1_vect){

    // motor encoder check
    if((PINB & (1 << PB5))==0) {
        ++num_pulses;

        if(num_pulses >= 16 * max_motor_turns){
            stop_motor();
            PORTD ^= (1 << PD7);
            num_pulses=0;
        }
    }

    // buttons check
    if((PINB & (1 << PB0))==0) {
        // down pressed
        _delay_ms(60); // debounce
        if((PINB & (1 << PB0))==0) {
            if((PINB & (1 << PB4))==0){
                // both pressed, change tuning
                selected_tuning = (selected_tuning + 1) % NUM_TUNINGS;
                selection_index = 0;
            } else {
                USART0_print("down pressed");
                if (selection_index == 0) {
                    selection_index = 5;
                } else {
                    --selection_index;
                }
                stop_motor();
            }
        }
    }

    if((PINB & (1 << PB4))==0) {
        // up pressed
        _delay_ms(60); // debounce
        if((PINB & (1 << PB4))==0) {
            if((PINB & (1 << PB0))==0){
                // both pressed, change tuning
                selected_tuning = (selected_tuning + 1) % NUM_TUNINGS;
                selection_index = 0;
            } else {
                USART0_print("up pressed");
                selection_index = (selection_index + 1) % 6;
            }
        }
        stop_motor();
    }

    if((PINB & (1 << PB6))==0) {
        // minus pressed
        _delay_ms(60); // debounce
        if((PINB & (1 << PB6))==0) {
            if ((PINB & (1 << PB7)) == 0) {
                //if both plus and minus pressed change directions
                reverse_directions = !reverse_directions;
            } else {
                USART0_print("minus pressed");
                USART0_print("minus button requested move");
                move_motor(BWD, 200, 15);
            }
        }
    }

    if((PINB & (1 << PB7))==0) {
        // plus pressed
        _delay_ms(60);
        if((PINB & (1 << PB7))==0) {
            if((PINB & (1 << PB6))==0){
                //if both plus and minus pressed change directions
                reverse_directions = !reverse_directions;
            } else {
                USART0_print("plus pressed");
                USART0_print("plus button requested move");
                move_motor(FWD, 200, 15);
            }
        }
    }
}

void motor_monitor_init(){
    // set up interrupts for the motor encoder
    DDRB &= ~(1 << PB5);
    PORTB |= (1<<PB5); //pullup

    PCICR |= (1 << PCIE1);
    PCMSK1 |= (1 << PCINT13);

    stop_motor(); // at the begining the motor is off
}

void buttons_init(){
    // PB0 = down
    // PB4 = up
    // PB6 = minus
    // PB7 = plus

    DDRB &= ~((1 << PB0) | (1 << PB4) | (1 << PB6) | (1 << PB7));

    PORTB |= (1<<PB0) | (1<<PB4) | (1<<PB6) | (1<<PB7); //pullup

    PCICR |= (1 << PCIE1);
    PCMSK1 |= (1 << PCINT8) | (1 << PCINT12) | (1 << PCINT14) | (1 << PCINT15);
}

int main() {

    motor_monitor_init();
    buttons_init();

    USART0_init();

    /* activate LCD backlight */

    DDRC |= (1 << PC2);
    PORTC |= (1 << PC2);

    LCD_init();
    LCD_waitNotBusy();

    /* activate PM board LED */

    DDRD |= (1 << PD7);
    PORTD |= (1 << PD7);

    // activate interrupts
    sei();

    init_adc();

    // init control motor
    init_PWM();

    char *strings = stringsE;
    char *description = descriptionE;

    for(;;){
        // main loop

        if(selected_tuning == 0) {
            selected_string = standardE[selection_index];
            strings = stringsE;
            description = descriptionE;
        } else if (selected_tuning == 1){
            selected_string = dropD[selection_index];
            strings = stringsDropD;
            description = descriptionDropD;
        } else if (selected_tuning == 2){
            selected_string = standardD[selection_index];
            strings = stringsStandardD;
            description = descriptioStandardD;
        } else {
            // default tuning is standard E
            selected_string = standardE[selection_index];
            strings = stringsE;
            description = descriptionE;
        }

        if (checkMaxAmp>ampThreshold){
            frequency = 38462/(float)(period);//calculate frequency timer rate/period
        } else {
            // if signal not strong enough do nothing
            stop_motor();
            frequency = 0;
        }

        if((frequency > (selected_string - (STRING_TOLERANCE * selected_string))) &&
                (frequency < (selected_string + (STRING_TOLERANCE * selected_string)))){

            // if measurement is in a reasonable range of +/- STRING_TOLERANCE %

            char buf[50];
            sprintf(buf, "%f\r\n", frequency);
            USART0_print(buf);

            LCD_clearEntireLine(1);

            if(frequency < selected_string - PRECISION){
                LCD_printAt(0x03, "<- LOW");
                USART0_print("LOW string requested move");
                move_motor(BWD, 200, (unsigned char)((selected_string - frequency) * multiplier[selection_index]));
            } else if (frequency > selected_string + PRECISION){
                LCD_printAt(0x06, "HIGH ->");
                USART0_print("HIGH string requested move");
                move_motor(FWD, 200, (unsigned char)((frequency - selected_string) * multiplier[selection_index]));
            } else {
                LCD_printAt(0x07, "OK");
                stop_motor();

                for(int j = 0; j < 4; ++j) {
                    _delay_ms(100);
                    PORTC ^= (1 << PC2);
                    _delay_ms(100);
                    PORTC ^= (1 << PC2);
                    _delay_ms(100);
                }

                LCD_clearEntireLine(1);

                // go to the next string if this one is OK

                LCD_printAt(0x03, description);
                selection_index = (selection_index + 1) % 6;
            }

        } else {
            LCD_clearEntireLine(1);
            LCD_printAt(0x03, description);
            stop_motor();
        }

        LCD_printAt(LCD_INSTR_nextLine, strings);

        LCD_writeInstr(0x80 + LCD_INSTR_nextLine + (3 * selection_index));

        _delay_ms(100);
    }

    return 0;
}

